# app.py
#
# Copyright 2022 johannesjh
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import gi

gi.require_version("Gdk", "4.0")
gi.require_version("Gtk", "4.0")
gi.require_version("Adw", "1")

from gi.repository import Adw, Gdk, Gio, Gtk

from .about import AboutDialog
from .window import FavagtkWindow


class FavagtkApplication(Adw.Application):
    """The main application singleton class."""

    def __init__(self):
        super().__init__(
            application_id="org.gnome.gitlab.johannesjh.favagtk",
            flags=Gio.ApplicationFlags.HANDLES_OPEN,
        )
        self.create_action("quit", self.do_quit, ["<primary>q"])
        self.create_action("about", self.on_about_action)

    def do_activate(self):
        """Called when the application is activated.

        We raise the application's main window, creating it if
        necessary.
        """
        self.load_css("/io/github/johannesjh/favagtk/app.css")
        win = self.props.active_window
        if not win:
            win = FavagtkWindow(application=self)
        win.present()

    def do_open(self, files, _n_files, _hint):
        """Called when the application shall open a file."""
        self.do_activate()
        win = self.props.active_window
        win.open_file(files[0])

    def load_css(self, resource_path):
        """Loads css from a Gio resource at given path"""
        display = Gdk.Display.get_default()
        assert display is not None
        provider = Gtk.CssProvider()
        provider.load_from_resource(resource_path)
        assert provider is not None
        Gtk.StyleContext.add_provider_for_display(
            display, provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )

    def on_about_action(self, widget, _):
        """Callback for the app.about action."""
        about = AboutDialog(parent=self.props.active_window)
        about.present()

    def create_action(self, name, callback, shortcuts=None):
        """Add an application action.

        Args:
            name: the name of the action
            callback: the function to be called when the action is
              activated
            shortcuts: an optional list of accelerators
        """
        action = Gio.SimpleAction.new(name, None)
        action.connect("activate", callback)
        self.add_action(action)
        if shortcuts:
            self.set_accels_for_action(f"app.{name}", shortcuts)

    def do_quit(self, *args):
        """
        Handler for the "quit" signal.
        Destroys all application windows.
        In consequence, GTK will terminate the application.
        """
        for win in self.get_windows():
            win.emit("close-request")
